#ifndef _CAR_H
#define _CAR_H

#include <string>
#include <ostream>
#include <list>
#include <map>

#include "cardynamics.h"
#include "scenegraph.h"
#include "model_joe03.h"
#include "texture.h"
#include "quaternion.h"
#include "mathvector.h"
#include "configfile.h"
#include "carinput.h"
#include "sound.h"
#include "camera.h"
#include "aabb.h"
#include "serialization.h"

class BEZIER;
class PERFORMANCE_TESTING;

class CAR
{
friend class PERFORMANCE_TESTING;
friend class SERIALIZER;
private:
	CARDYNAMICS <double> dynamics;

	SCENENODE * topnode;

	DRAWABLE * bodydraw;
	DRAWABLE * interiordraw;
	DRAWABLE * glassdraw;
	SCENENODE * bodynode;
	MODEL_JOE03 bodymodel;
	MODEL_JOE03 interiormodel;
	MODEL_JOE03 glassmodel;
	TEXTURE_GL bodytexture;
	TEXTURE_GL bodytexture_misc1;
	TEXTURE_GL interiortexture;
	TEXTURE_GL interiortexture_misc1;
	TEXTURE_GL glasstexture;
	TEXTURE_GL glasstexture_misc1;
	MODEL_JOE03 collisionmodel;
	DRAWABLE * driverdraw;
	SCENENODE * drivernode;
	MODEL_JOE03 drivermodel;
	TEXTURE_GL drivertexture;
	TEXTURE_GL drivertexture_misc1;

	std::map <std::string, SOUNDBUFFER> soundbuffers;
	class ENGINESOUNDINFO
	{
		public:
			float minrpm, maxrpm, naturalrpm, fullgainrpmstart, fullgainrpmend;
			enum
			{
				POWERON,
    				POWEROFF,
				BOTH
			} power;

			ENGINESOUNDINFO() : minrpm(1.0), maxrpm(100000.0), naturalrpm(7000.0),fullgainrpmstart(minrpm),fullgainrpmend(maxrpm),power(BOTH) {}

			bool operator < (const ENGINESOUNDINFO & other) const {return minrpm < other.minrpm;}
	};
	std::list <std::pair <ENGINESOUNDINFO, SOUNDSOURCE> > enginesounds;

	DRAWABLE * wheeldraw[4];
	SCENENODE * wheelnode[4];
	DRAWABLE * debugwheeldraw[40]; //10 debug wheels per wheel
	SCENENODE * debugwheelnode[40]; //10 debug wheels per wheel
	MODEL_JOE03 wheelmodelfront;
	MODEL_JOE03 wheelmodelrear;
	TEXTURE_GL wheeltexturefront;
	TEXTURE_GL wheeltexturerear;
	TEXTURE_GL wheeltexturefront_misc1;
	TEXTURE_GL wheeltexturerear_misc1;
	TEXTURE_GL braketexture;

	SOUNDSOURCE tiresqueal[4];
	SOUNDSOURCE roadnoise;

	MATHVECTOR <float, 3> view_position;
	MATHVECTOR <float, 3> hood_position;

	//CAMERA_FIXED driver_cam;
	CAMERA_MOUNT driver_cam;
	//CAMERA_FIXED hood_cam;
	CAMERA_MOUNT hood_cam;
	CAMERA_CHASE rigidchase_cam;
	CAMERA_CHASE chase_cam;
	CAMERA_ORBIT orbit_cam;

	//internal variables that might change during driving (so, they need to be serialized)
	bool auto_clutch;
	float last_auto_clutch;
	float remaining_shift_time;
	float shift_clutch_start;
	int shift_gear;
	bool shifted;
	float last_steer;
	bool autoshift_enabled;
	float shift_time;

	const bool debug_wheel_draw;

	std::string cartype;
	int sector; //the last lap timing sector that the car hit
	const BEZIER * curpatch[4]; //the last bezier patch that each wheel hit

	///take the parentnode, add a scenenode (only if output_scenenodeptr is NULL), add a drawable to the scenenode, load a model, load a texture, and set up the drawable with the model and texture.
	/// returns true if successful
	bool LoadInto(SCENENODE * parentnode, SCENENODE * & output_scenenodeptr, DRAWABLE * & output_drawableptr, const std::string & joefile,
		      MODEL_JOE03 & output_model, const std::string & texfile, TEXTURE_GL & output_texture_diffuse,
		const std::string & misc1texfile, TEXTURE_GL & output_texture_misc1,
  		int anisotropy, const std::string & texsize, std::ostream & error_output);

	void CopyPhysicsResultsIntoDisplay();

	template <typename T>
	bool GetConfigfileParam(std::ostream & error_output, CONFIGFILE & c, const std::string & paramname, T & output) const
	{
		if (!c.GetParam(paramname, output))
		{
			error_output << "Couldn't get car parameter: " << paramname << std::endl;
			return false;
		}
		else
			return true;
	}

	bool LoadDynamics(CONFIGFILE & c, std::ostream & error_output);

	float AutoClutch(float last_clutch, float dt) const;
	float ShiftAutoClutch() const;
	float ShiftAutoClutchThrottle() const; ///< return the maximum throttle position taking into account autoclutch work due to shifting

	int AutoShift() const;

	float DownshiftPoint(int gear) const;

	void ShiftGears(int new_gear)
	{
		//remaining_shift_time = 0.3;
		remaining_shift_time = shift_time;
		shift_gear = new_gear;
		shifted = false;
		shift_clutch_start = dynamics.GetClutch().GetClutch();
	}

	void UpdateSounds(float dt);
	void UpdateCameras(float dt);

	bool LoadSounds(const std::string & carpath, const std::string & carname, const SOUNDINFO & sound_device_info, const SOUNDBUFFERLIBRARY & soundbufferlibrary, std::ostream & info_output, std::ostream & error_output);

public:
	CAR() : topnode(NULL),bodydraw(NULL),bodynode(NULL),drivernode(NULL),auto_clutch(true),last_auto_clutch(1.0),remaining_shift_time(0.0),shift_gear(0),shifted(true),last_steer(0),autoshift_enabled(false),shift_time(0.2),debug_wheel_draw(false),sector(-1)
	{
		for (int i = 0; i < 4; i++)
		{
			curpatch[i] = NULL;
			wheelnode[i] = NULL;
		}
	}
	~CAR();
	bool Load(CONFIGFILE & carconf, const std::string & carpath, const std::string & driverpath, const std::string & carname, const std::string & carpaint,
		  const MATHVECTOR <float, 3> & initial_position, const QUATERNION <float> & initial_orientation,
    		  SCENENODE & sceneroot, bool soundenabled, const SOUNDINFO & sound_device_info, const SOUNDBUFFERLIBRARY & soundbufferlibrary,
		  int anisotropy, bool defaultabs, bool defaulttcs, const std::string & texsize, float camerabounce,
    		  std::ostream & info_output, std::ostream & error_output);

	void Tick(double dt)
	{
		remaining_shift_time-=dt;
		if (remaining_shift_time < 0)
			remaining_shift_time = 0;
		dynamics.Tick(dt);
		CopyPhysicsResultsIntoDisplay();
		UpdateSounds(dt);
		UpdateCameras(dt);
	}

	bool Shifting() const
	{
		return (remaining_shift_time > 0);
	}

	void GetSoundList(std::list <SOUNDSOURCE *> & outputlist)
	{
		for (std::list <std::pair <ENGINESOUNDINFO, SOUNDSOURCE> >::iterator i = enginesounds.begin(); i != enginesounds.end(); i++)
		{
			outputlist.push_back(&i->second);
		}

		for (int i = 0; i < 4; i++)
			outputlist.push_back(&tiresqueal[i]);

		outputlist.push_back(&roadnoise);
	}

	void GetEngineSoundList(std::list <SOUNDSOURCE *> & outputlist)
	{
		for (std::list <std::pair <ENGINESOUNDINFO, SOUNDSOURCE> >::iterator i = enginesounds.begin(); i != enginesounds.end(); i++)
		{
			outputlist.push_back(&i->second);
		}
	}

	///returns the position of the center of the wheel
	const MATHVECTOR <float, 3> GetWheelPosition(const WHEEL_POSITION wpos) const
	{
		MATHVECTOR <float, 3> v;
		v=dynamics.GetWheelPosition(wpos);
		return v;
	}

	///returns the position of the center of the wheel at the given suspension displacement
	const MATHVECTOR <float, 3> GetWheelPositionAtDisplacement(const WHEEL_POSITION wpos, const float displacement) const
	{
		MATHVECTOR <float, 3> v;
		v=dynamics.GetWheelPositionAtDisplacement(wpos, displacement);
		return v;
	}

	float GetTireRadius(const WHEEL_POSITION wpos) const
	{
		return dynamics.GetTire(wpos).GetRadius();
	}

	const MATHVECTOR <float, 3> GetDownVector() const
	{
		MATHVECTOR <float, 3> v;
		v.Set(0,0,-1);
		dynamics.GetBody().GetOrientation().RotateVector(v);
		return v;
	}

	void SetWheelContactProperties(WHEEL_POSITION wheel_index, float wheelheight, const MATHVECTOR <float, 3> & position, const MATHVECTOR <float, 3> & normal,
				       float bumpwave, float bumpamp, float frict_no, float frict_tread, float roll_res, float roll_drag)
	{
		//std::cout << wheel_index << ", " << wheelheight << std::endl;
		MATHVECTOR <double, 3> dblpos;
		dblpos = position;
		MATHVECTOR <double, 3> dblnorm;
		dblnorm = normal;
		dynamics.SetWheelContactProperties(wheel_index, wheelheight, dblpos, dblnorm,
						  bumpwave, bumpamp, frict_no, frict_tread,
						  roll_res, roll_drag);
	}

	void HandleInputs(const std::vector <float> & inputs, float dt);

	int GetEngineRPM() const
	{
		return dynamics.GetEngine().GetRPM();
	}

	int GetEngineStallRPM() const
	{
		return dynamics.GetEngine().GetStallRPM();
	}

	int GetEngineRedline() const
	{
		return dynamics.GetEngine().GetRedline();
	}

	int GetEngineRPMLimit() const
	{
		return dynamics.GetEngine().GetRPMLimit();
	}

	int GetGear() const
	{
		return dynamics.GetTransmission().GetGear();
	}

	float GetClutch() const
	{
		return dynamics.GetClutch().GetClutch();
	}

	void DebugPrint(std::ostream & out, bool p1, bool p2, bool p3, bool p4)
	{
		dynamics.DebugPrint(out, p1, p2, p3, p4);
	}

	void SetAutoClutch(bool new_auto_clutch)
	{
		auto_clutch = new_auto_clutch;
	}

	///set the position of the center of mass of the car
	void SetPosition(const MATHVECTOR <float,3> & new_position)
	{
		MATHVECTOR <double,3> newpos;
		newpos = new_position;
		dynamics.GetBody().SetPosition(newpos);

		QUATERNION <float> floatq;
		floatq = dynamics.GetBody().GetOrientation();
		chase_cam.Reset(GetCenterOfMassPosition(), floatq);

		MATHVECTOR <double, 3> doublevec;
		doublevec = view_position;
		MATHVECTOR <float, 3> floatvec;
		floatvec = dynamics.CarLocalToWorld(doublevec);
		driver_cam.Reset(floatvec, floatq);
		doublevec = hood_position;
		floatvec = dynamics.CarLocalToWorld(doublevec);
		hood_cam.Reset(floatvec, floatq);
	}

	MATHVECTOR <float, 3> GetCenterOfMassPosition() const
	{
		MATHVECTOR <float,3> pos;
		pos = dynamics.GetBody().GetPosition();
		return pos;
	}

	MATHVECTOR <float, 3> GetOriginPosition() const
	{
		MATHVECTOR <float,3> pos;
		pos = dynamics.GetOriginPosition();
		return pos;
	}

	QUATERNION <float> GetOrientation() const
	{
		QUATERNION <float> q;
		q = dynamics.GetBody().GetOrientation();
		return q;
	}

	const CAMERA * GetDriverCamera() const {return &driver_cam;}
	const CAMERA * GetHoodCamera() const {return &hood_cam;}
	const CAMERA * GetOrbitCamera() const {return &orbit_cam;}
	const CAMERA * GetChaseCamera() const {return &chase_cam;}
	const CAMERA * GetRigidChaseCamera() const {return &rigidchase_cam;}

	const AABB <float> GetCollisionDimensions()
	{
		return collisionmodel.GetAABB();
	}

	void SetChassisContact(const MATHVECTOR <float, 3> & pos, const MATHVECTOR <float, 3> & normal, float depth, float dt)
	{
		if (depth > 0.0)
		{
			/*MATHVECTOR <double, 3> translation;
			translation = normal*(depth + 0.00);
			dynamics.Translate (translation);*/

			MATHVECTOR <double, 3> dpos;
			dpos = pos;
			MATHVECTOR <double, 3> dnormal;
			dnormal = normal;
			dynamics.ProcessContact(dpos, dnormal, depth, dt);
		}
	}

	///return the speedometer reading (based on the driveshaft speed) in m/s
	float GetSpeedometer()
	{
		return dynamics.GetSpeedo();
	}

	std::string GetCarType() const
	{
		return cartype;
	}

	void SetSector ( int value )
	{
		sector = value;
	}

	int GetSector() const
	{
		return sector;
	}

	void SetCurPatch (unsigned int wheel, const BEZIER* value )
	{
		assert (wheel < 4);
		curpatch[wheel] = value;
	}

	const BEZIER* GetCurPatch(unsigned int wheel) const
	{
		assert (wheel < 4);
		return curpatch[wheel];
	}

	bool GetABSEnabled() const {return dynamics.GetABSEnabled();}
	bool GetABSActive() const {return dynamics.GetABSActive();}
	bool GetTCSEnabled() const {return dynamics.GetTCSEnabled();}
	bool GetTCSActive() const {return dynamics.GetTCSActive();}

	float GetLastSteer() const
	{
		return last_steer;
	}

	float GetSpeed()
	{
		return dynamics.GetSpeed();
	}

	void SetAutoShift ( bool value )
	{
		autoshift_enabled = value;
	}

	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s,dynamics);
		_SERIALIZE_(s,last_auto_clutch);
		_SERIALIZE_(s,remaining_shift_time);
		_SERIALIZE_(s,shift_clutch_start);
		_SERIALIZE_(s,shift_gear);
		_SERIALIZE_(s,shifted);
		_SERIALIZE_(s,last_steer);
		_SERIALIZE_(s,autoshift_enabled);
	}

	float GetFeedback()
	{
		return 0.5*(dynamics.GetTire(FRONT_LEFT).GetFeedback()+dynamics.GetTire(FRONT_RIGHT).GetFeedback());
	}

	///returns a float from 0.0 to 1.0 with the amount of tire squealing going on
	float GetTireSquealAmount(WHEEL_POSITION i, float dt) const
	{
		MATHVECTOR <float, 3> groundvel;
		groundvel = dynamics.GetWheelVelocity(WHEEL_POSITION(i), dt);
		QUATERNION <float> wheelspace;
		wheelspace = dynamics.GetBody().GetOrientation() * dynamics.GetWheelSteeringAndSuspensionOrientation(WHEEL_POSITION(i));
		(-wheelspace).RotateVector(groundvel);
		float wheelspeed = dynamics.GetWheel(WHEEL_POSITION(i)).GetAngularVelocity()*dynamics.GetTire(WHEEL_POSITION(i)).GetRadius();
		groundvel[0] -= wheelspeed;
		groundvel[2] = 0;
		float squeal = (groundvel.Magnitude()-5.0)*0.2;
		if (squeal < 0)
			squeal = 0;
		if (squeal > 1)
			squeal = 1;

		return squeal;
	}

	///used by the AI
	float GetAerodynamicDownforceCoefficient() const
	{
		return dynamics.GetAerodynamicDownforceCoefficient();
	}

	///used by the AI
	float GetAeordynamicDragCoefficient() const
	{
		return dynamics.GetAeordynamicDragCoefficient();
	}

	float GetMass() const
	{
		return dynamics.GetBody().GetMass();
	}

	MATHVECTOR <float, 3> GetVelocity() const
	{
		MATHVECTOR <float, 3> vel;
		vel = dynamics.GetBody().GetVelocity();
		return vel;
	}

	///used by AI
	float GetTireMaxFx(WHEEL_POSITION tire_index) const
	{
		return dynamics.GetTire(tire_index).GetMaximumFx(GetMass()*0.25*9.81);
	}

	///used by AI
	float GetTireMaxFy(WHEEL_POSITION tire_index) const
	{
		return dynamics.GetTire(tire_index).GetMaximumFy(GetMass()*0.25*9.81, 0.0);
	}

	///Get the maximum steering angle in degrees.  used by the AI
	float GetMaxSteeringAngle() const
	{
		return dynamics.GetMaxSteeringAngle();
	}
};

#endif
