#ifndef _GAME_H
#define _GAME_H

#include <ostream>
#include <string>
#include <list>
#include <map>
#include <vector>
#include <memory>

#include "graphics.h"
#include "eventsystem.h"
#include "settings.h"
#include "pathmanager.h"
#include "track.h"
#include "mathvector.h"
#include "quaternion.h"
#include "font.h"
#include "text_draw.h"
#include "gui.h"
#include "car.h"
#include "collision_detection.h"
#include "carcontrolmap_local.h"
#include "hud.h"
#include "game_render.h"
#include "inputgraph.h"
#include "sound.h"
#include "camera.h"
#include "trackmap.h"
#include "loadingscreen.h"
#include "timer.h"
#include "replay.h"
#include "forcefeedback.h"
#include "particle.h"
#include "ai.h"

class GAME
{
private:
	float TickPeriod() const {return framerate;}
	void MainLoop();
	bool ParseArguments(std::list <std::string> & args);
	void InitializeCoreSubsystems();
	void InitializeThreading();
	bool InitializeSound();
	bool InitializeGUI();
	void End();
	void Test();
	void Tick(float dt);
	void Draw();
	void AdvanceGameLogic();
	void UpdateCar(CAR & car);
	void UpdateCarInputs(CAR & car);
	void UpdateCarCollisions(CAR & car);
	void UpdateTimer();
	void ProcessGUIInputs();
	void ProcessGameInputs();
	void ProcessGUIAction(const std::string & action);
	bool NewGame(bool playreplay=false, bool opponents=false, int num_laps=0);
	void LeaveGame();
	bool LoadTrack(const std::string & trackname);
	bool LoadCar(const std::string & carname, const std::string & carpaint, const MATHVECTOR <float, 3> & start_position, const QUATERNION <float> & start_orientation, bool islocal, bool isai, const std::string & carfile=""); ///< carfile is a string containing an entire .car file (e.g. XS.car) and is used instead of reading from disk.  this is optional
	bool AlignCarsWithGround();
	bool LoadFonts();
	void CalculateFPS();
	void PopulateValueLists(std::map<std::string, std::list <std::pair<std::string,std::string> > > & valuelists);
	void PopulateReplayList(std::list <std::pair <std::string, std::string> > & replaylist);
	void PopulateCarPaintList(const std::string & carname, std::list <std::pair <std::string, std::string> > & carpaintlist);
	enum OPTION_ACTION
	{
		SAVE,
		LOAD
	};
	void LoadSaveOptions(OPTION_ACTION action, std::map<std::string, std::string> & options);
	void CollapseSceneToDrawlistmap(SCENENODE & node, std::map < DRAWABLE_FILTER *, std::vector <SCENEDRAW> > & outputmap);
	void UpdateTrackMap();
	void LoadingScreen(float progress, float max);
	void ProcessNewSettings();
	void RedisplayControlPage();
	void LoadControlsIntoGUIPage(const std::string & pagename);
	void UpdateForceFeedback(float dt);
	void UpdateParticleSystems(float dt);
	void AddTireSmokeParticles(float dt, CAR & car);

	std::ostream & info_output;
	std::ostream & error_output;
	unsigned int frame;
	double target_time;
	const float framerate;
	const float collision_rate;
	unsigned int collision_frameskip;
	unsigned int cur_collision_frameskip;
	TEXT_DRAW fps;
	DRAWABLE * fps_draw;
	std::vector <float> fps_track;
	int fps_position;
	bool multithreaded;
	GAME_RENDER rendertask;
	const CAMERA * active_camera;
	CAMERA_FREE free_camera;
	bool pause;
	unsigned int particle_timer;
	bool maximize_parallel_render;
	std::vector <std::pair<std::string, std::string> > opponents; //pairs of car names and car paints for opponents
	int race_laps;

	std::string controlgrab_page;
	std::string controlgrab_input;
	bool controlgrab_analog;
	bool controlgrab_only_one;
	std::pair <int,int> controlgrab_mouse_coords;
	CARCONTROLMAP_LOCAL::CONTROL controlgrab_editcontrol;
	std::vector <EVENTSYSTEM_SDL::JOYSTICK> controlgrab_joystick_state;

	GRAPHICS_SDLGL graphics;
	EVENTSYSTEM_SDL eventsystem;
	SOUND sound;
	SOUNDBUFFERLIBRARY generic_sounds;
	SETTINGS settings;
	PATHMANAGER pathmanager;
	TRACK track;
	TRACKMAP trackmap;
	GUI gui;
	std::map <std::string, FONT> fonts;
	SCENENODE rootnode;
	SCENENODE * tracknode;
	std::list <CAR> cars;
	std::map <CAR *, int> cartimerids;
	std::pair <CAR *, CARCONTROLMAP_LOCAL> carcontrols_local;
	std::map <CAR *, std::vector <COLLISION_CONTACT> > cached_collisions;
	COLLISION_WORLD collision;
	HUD hud;
	INPUTGRAPH inputgraph;
	LOADINGSCREEN loadingscreen;
	SCENENODE loadingscreen_node;
	TIMER timer;
	REPLAY replay;
	PARTICLE_SYSTEM tire_smoke;
	AI ai;

#ifdef ENABLE_FORCE_FEEDBACK
	std::auto_ptr <FORCEFEEDBACK> forcefeedback;
	double ff_update_time;
#endif

public:
	GAME(std::ostream & info_out, std::ostream & err_out) :
		info_output(info_out), error_output(err_out), frame(0),
		target_time(0),framerate(0.004),collision_rate(0.02),collision_frameskip(collision_rate/framerate),
		cur_collision_frameskip(0), fps_draw(NULL),
		fps_track(10,0),fps_position(0),
		multithreaded(false), active_camera(NULL), pause(false), particle_timer(0),
		maximize_parallel_render(false), race_laps(0), tracknode(NULL), replay(framerate){carcontrols_local.first = NULL;}

	void Start(std::list <std::string> & args);
};

#endif
