#ifndef _SERIALIZATION_H
#define _SERIALIZATION_H

#include <list>
#include <vector>
#include <map>
#include <iostream>
#include <string>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <iomanip>
#include <cstring>
#include <cassert>

#include "simpletree.h"

using std::cout;
using std::cerr;
using std::endl;
using std::list;
using std::vector;
using std::map;
using std::string;
using std::stringstream;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
using std::pair;

#define _GENERATEUID_(ptr,uidoutput) unsigned long long uidoutput = (size_t)ptr

class SERIALIZER
{
private:
	list < pair <void **,unsigned long long> > pointers_to_be_resolved;
	map <unsigned long long, void *> pointer_resolution_map;
	bool trackpointers;

protected:
	enum FORMAT_TYPE
	{
		START,
  		END
	};
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin) = 0;
	virtual void SerializeData(int & t) = 0;
	virtual void SerializeData(float & t) = 0;
	virtual void SerializeData(double & t) = 0;
	virtual void SerializeData(string & t) = 0;
	virtual void SerializeData(unsigned long long & t) = 0;
	bool TrackingPointers() {return trackpointers;}

public:
	SERIALIZER() : trackpointers(true) {Reset();}
	virtual ~SERIALIZER() {}
	void Reset() {pointers_to_be_resolved.clear();pointer_resolution_map[0]=0;}
	void SetPointerTracking(bool newtrack) {trackpointers = newtrack;}

	virtual bool IsOutput() = 0;

	template <typename T>
	void StartSerialize(string name, T & t)
	{
		Serialize(name, t);
		if (!this->IsOutput() && !pointers_to_be_resolved.empty() && TrackingPointers())
		{
			ResolvePointers();
		}
	}

	//do a nested serialize when the type is not built-in
	template <typename T>
	void Serialize(string name, T & t)
	{
		_GENERATEUID_(&t,uid);
		unsigned long long fileuid = this->FormatInfo(name, START, uid, false);
		if (!this->IsOutput())
			pointer_resolution_map[fileuid] = &t;
		t.Serialize(this);
		this->FormatInfo(name, END, uid, false);
	}

	template <typename U>
	void SerializeList(string name, list <U> & t)
	{
		_GENERATEUID_(&t,uid);

		unsigned long long fileuid = this->FormatInfo(name, START, uid, false);
		if (!this->IsOutput())
			pointer_resolution_map[fileuid] = &t;

		if (this->IsOutput())
		{
			int listsize = t.size();

			//this->SerializeData(listsize);
            this->FormatInfo("size", START, 0, true);
            this->SerializeData(listsize);
            this->FormatInfo("size", END, uid, true);

			int count = 1;
			for (typename list <U>::iterator i = t.begin(); i != t.end(); i++,count++)
			{
				stringstream itemname;
				itemname << "Item" << count;
				this->Serialize(itemname.str(), *i);
			}
		}
		else
		{
			t.clear();

			int listsize;
			//this->SerializeData(listsize);
			this->FormatInfo("size", START, 0, true);
            this->SerializeData(listsize);
            this->FormatInfo("size", END, uid, true);

			for (int i = 0; i < listsize; i++)
			{
				stringstream itemname;
				itemname << "Item" << i+1;
				U tempobj;
				t.push_back(tempobj);
				this->Serialize(itemname.str(), t.back());
			}
		}

		this->FormatInfo(name, END, uid, false);
	}
	template <typename U>
	void SerializeVector(string name, vector <U> & t)
	{
		_GENERATEUID_(&t,uid);

		unsigned long long fileuid = this->FormatInfo(name, START, uid, false);
		if (!this->IsOutput())
			pointer_resolution_map[fileuid] = &t;

		if (this->IsOutput())
		{
			int listsize = t.size();

			//this->SerializeData(listsize);
			this->FormatInfo("size", START, 0, true);
            this->SerializeData(listsize);
            this->FormatInfo("size", END, uid, true);

			int count = 1;
			for (typename vector <U>::iterator i = t.begin(); i != t.end(); i++,count++)
			{
				stringstream itemname;
				itemname << "Item" << count;
				this->Serialize(itemname.str(), *i);
			}
		}
		else
		{
			//t.clear();
			int listsize;
			//this->SerializeData(listsize);
			this->FormatInfo("size", START, 0, true);
            this->SerializeData(listsize);
            this->FormatInfo("size", END, uid, true);

			t.resize(listsize);
			for (int i = 0; i < listsize; i++)
			{
				stringstream itemname;
				itemname << "Item" << i+1;
				//U tempobj;
				//t.push_back(tempobj);
				this->Serialize(itemname.str(), t[i]);
			}
		}

		this->FormatInfo(name, END, uid, false);
	}
	template <typename U, typename T>
	void SerializeMap(string name, map <U, T> & t)
	{
		_GENERATEUID_(&t,uid);

		unsigned long long fileuid = this->FormatInfo(name, START, uid, false);
		if (!this->IsOutput())
			pointer_resolution_map[fileuid] = &t;

		if (this->IsOutput())
		{
			int mapsize = t.size();
			//this->SerializeData(mapsize);
			this->FormatInfo("size", START, 0, true);
            this->SerializeData(mapsize);
            this->FormatInfo("size", END, uid, true);

			int count = 1;
			for (typename map <U,T>::iterator i = t.begin(); i != t.end(); i++,count++)
			{
				stringstream countstr;
				countstr << count;

				string keystr = "Key" + countstr.str();
				string valstr = "Value" + countstr.str();

				U tempkey = i->first;
				this->Serialize(keystr, tempkey);
				this->Serialize(valstr, i->second);
			}
		}
		else
		{
			t.clear();
			int mapsize;
			//this->SerializeData(mapsize);
			this->FormatInfo("size", START, 0, true);
            this->SerializeData(mapsize);
            this->FormatInfo("size", END, uid, true);

			for (int i = 0; i < mapsize; i++)
			{
				stringstream countstr;
				countstr << i+1;

				string keystr = "Key" + countstr.str();
				string valstr = "Value" + countstr.str();

				U tempkey;
				this->Serialize(keystr, tempkey);

				//T tempval;
				//this->Serialize(valstr, tempval);
				//t[tempkey] = tempval;
				this->Serialize(valstr, t[tempkey]);
			}
		}

		this->FormatInfo(name, END, uid, false);
	}
	template <typename T>
	void SerializePointer(string name, T * & t)
	{
		_GENERATEUID_(&t,uid);
		unsigned long long fileuid = this->FormatInfo(name, START, uid, true);

		if (!TrackingPointers())
		{
			unsigned long long ptr_uid = 0;
			this->SerializeData(ptr_uid);
		}
		else
		{
			if (!this->IsOutput())
				pointer_resolution_map[fileuid] = &t;

			unsigned long long ptr_uid = (size_t) t;
			this->SerializeData(ptr_uid);

			if (!this->IsOutput())
			{
				T ** tdouble = &t;
				void ** tvoid = reinterpret_cast<void **> (tdouble);
				pointers_to_be_resolved.push_back(pair <void **, unsigned long long> (tvoid,ptr_uid));
				t = NULL;
			}

			this->FormatInfo(name, END, uid, true);
		}
	}

	void ResolvePointers()
	{
		for (list < pair <void **,unsigned long long> >::iterator i = pointers_to_be_resolved.begin();
				   i != pointers_to_be_resolved.end(); i++)
		{
			if (pointer_resolution_map.find(i->second) == pointer_resolution_map.end())
			{
				//uh oh, throw some kind of error
				cerr << "Serialization Error: Couldn't resolve pointer with uid " << i->second << endl;
			}
			else
			{
				*(i->first) = pointer_resolution_map[i->second];
			}
		}
	}
};

//serialization specializations
template <>
void SERIALIZER::Serialize<int>(string name, int & t);
template <>
void SERIALIZER::Serialize<bool>(string name, bool & t);
template <>
void SERIALIZER::Serialize<float>(string name, float & t);
template <>
void SERIALIZER::Serialize<double>(string name, double & t);
template <>
void SERIALIZER::Serialize<string>(string name, string & t);

//implements XML formatting
class SERIALIZER_XML_OUTPUT : public SERIALIZER
{
private:
	ostream * out;

protected:
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
		if (type == START)
		{
			if (TrackingPointers())
				*out << "<" << name << " uid=\"" << uid << "\">" << endl;
			else
				*out << "<" << name << ">" << endl;
		}
		else if (type == END)
			*out << "</" << name << ">" << endl;
		return uid;
	}
	virtual void SerializeData(int & i) {*out << i << endl;}
	virtual void SerializeData(float & i) {*out << std::scientific << std::setprecision (20) << i << endl;}
	virtual void SerializeData(double & i) {*out << std::scientific << std::setprecision (20) << i << endl;}
	virtual void SerializeData(string & i) {*out << i << endl;}
	virtual void SerializeData(unsigned long long & i) {*out << i << endl;}
public:
	SERIALIZER_XML_OUTPUT() : out(NULL) {}
	SERIALIZER_XML_OUTPUT(ostream & newout) : out(&newout) {}

	void SetOutput(ostream & outp) {out = &outp;}

	virtual bool IsOutput() {return true;}
};

class SERIALIZER_XML_INPUT : public SERIALIZER
{
private:
	istream * in;

	string dataread;

	string SeekTo(char delim)
	{
		string text;
		std::getline(*in,text,delim);
		return text;
	}

protected:
	virtual void SerializeData(int & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(float & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(double & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(string & i)
	{
		int start = 0;
		int end = dataread.size();
		if (dataread[0] == '\n')
			start = 1;
		if (dataread[end-1] == '\n')
			end--;

		i = dataread.substr(start,end-start);
	}
	virtual void SerializeData(unsigned long long & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
		if (type == END)
			SeekTo('/');

		stringstream rawstream(SeekTo('>'));
		//cerr << name << "," << uid << ": " << rawstream.str() << endl;

		string inname;
		rawstream >> inname;
		unsigned long long output_uid = uid;
		if (type == START && TrackingPointers())
		{
			string uidstring;
			std::getline(rawstream,uidstring,'"');
			std::getline(rawstream,uidstring,'"');
			stringstream uidstream(uidstring);
			uidstream >> output_uid;
		}

		//if ((type == START && name != inname) || (type == END && name != inname))
		if (name != inname)
		{
			//handle an error
			cerr << "Serialization Error: expected " << name << ", got " << inname << endl;
		}

		dataread = SeekTo('<');

		return output_uid;
	}
public:
	SERIALIZER_XML_INPUT() : in(NULL) {}
	SERIALIZER_XML_INPUT(istream & newin) : in(&newin) {SeekTo('<');}

	void SetInput(istream & inp) {in = &inp;SeekTo('<');}

	virtual bool IsOutput() {return false;}
};

//implements config file looking formatting
class SERIALIZER_CONFIGFILE_OUTPUT : public SERIALIZER
{
private:
	ostream * out;
	int indent;

	void PrintIndent()
	{
	    for (int i = 0; i < indent; i++)
            *out << "  ";
	}

	///replace <CR> with \n
	string Escape(const string & input)
	{
	    string outputstr;
	    for (unsigned int i = 0; i < input.length(); i++)
	    {
	        if (input[i] == '\n')
                outputstr.append("\\n");
            else
                outputstr.append(1,input[i]);
	    }
	    return outputstr;
	}

protected:
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
	    if (type == START)
		{
		    if (!builtin)
                indent++;

		    PrintIndent();

		    if (builtin)
                *out << name;
            else
                *out << "[ " << name << " ]";

			if (TrackingPointers())
				*out << " (" << uid << ")";

            if (builtin)
                *out << " = ";
            else
                *out << endl;
		}
		else if (type == END && !builtin)
		{
		    indent--;
		}

		return uid;
	}
	virtual void SerializeData(int & i) {*out << i << endl;}
	virtual void SerializeData(float & i) {*out << std::scientific << std::setprecision (20) << i << endl;}
	virtual void SerializeData(double & i) {*out << std::scientific << std::setprecision (20) << i << endl;}
	virtual void SerializeData(string & i) {*out << Escape(i) << endl;}
	virtual void SerializeData(unsigned long long & i) {*out << i << endl;}
public:
	SERIALIZER_CONFIGFILE_OUTPUT() : out(NULL),indent(-1) {}
	SERIALIZER_CONFIGFILE_OUTPUT(ostream & newout) : out(&newout),indent(-1) {}

	void SetOutput(ostream & outp) {out = &outp;}

	virtual bool IsOutput() {return true;}
};

class SERIALIZER_CONFIGFILE_INPUT : public SERIALIZER
{
private:
	istream * in;

	string dataread;

    /*///seeks to whichever delimiter is encountered first and returns the text minus the delimiter
    //BUGGY: DOESN'T HANDLE EOF CHAR
    string SeekTo(char delim1, char delim2)
    {
        string output;
        while (*in)
        {
            char newchar = in->get();
            if (newchar == delim1 || newchar == delim2)
                return output;
            else
                output.append(1,newchar);
        }
        return output;
    }*/

	string SeekTo(char delim)
	{
		string text;
		std::getline(*in,text,delim);
		return text;
	}

	string strLTrim(string instr)
    {
        return instr.erase(0, instr.find_first_not_of(" \t"));
    }

    string strRTrim(string instr)
    {
        if (instr.find_last_not_of(" \t") != string::npos)
            return instr.erase(instr.find_last_not_of(" \t") + 1);
        else
            return instr;
    }

    string strTrim(string instr)
    {
        return strLTrim(strRTrim(instr));
    }

    ///replace \n with <CR>
	string UnEscape(const string & input)
	{
	    string outputstr;
	    for (unsigned int i = 0; i < input.length(); i ++)
	    {
	        if ((unsigned int)(i + 1) == input.length())
                outputstr.append(1,input[i]);
            else
            {
                if (input[i] == '\\' && input[i+1] == 'n')
                {
                    outputstr.append(1,'\n');
                    i++;
                }
                else
                {
                    outputstr.append(1,input[i]);
                }
            }
	    }
	    return outputstr;
	}

protected:
	virtual void SerializeData(int & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(float & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(double & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual void SerializeData(string & i)
	{
	    i = strTrim(UnEscape(dataread));
	    //i = strTrim((dataread));
	    //std::cout << "original: \"" << dataread << "\"\n";
	    //std::cout << "i: \"" << i << "\"\n";
	}
	virtual void SerializeData(unsigned long long & i)
	{
		stringstream rawstream(dataread);
		rawstream >> i;
	}
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
	    unsigned long long output_uid = uid;

        if (type == START)
	    {
	        string incoming;
	        if (builtin)
                incoming = SeekTo('=');
            else
            {
                incoming = SeekTo('\n');
            }
	        stringstream rawstream(incoming);
	        string inname;
	        if (builtin)
                rawstream >> inname;
            else
            {
                std::getline(rawstream,inname,'[');
                std::getline(rawstream,inname,']');
                inname = strTrim(inname);
            }
	        //inname = strTrim(inname);

	        if (TrackingPointers())
            {
                string uidstring;
                std::getline(rawstream,uidstring,'(');
                std::getline(rawstream,uidstring,')');
                stringstream uidstream(uidstring);
                uidstream >> output_uid;
            }

            if (builtin)
                dataread = SeekTo('\n'); //put the rest of the line into dataread

	        if (name != inname)
            {
                //handle an error
                cerr << "Serialization Error: expected " << name << ", got " << inname << endl;
                cerr << "Incoming: \"" << incoming << "\"" << endl;
                cerr << builtin << endl;
                cerr << *in << endl;
                if (*in)
                {
                    cerr << SeekTo('\n') << endl;
                }
                assert(0);
            }
	    }

	    return output_uid;
	}
public:
	SERIALIZER_CONFIGFILE_INPUT() : in(NULL) {}
	SERIALIZER_CONFIGFILE_INPUT(istream & newin) : in(&newin) {}

	void SetInput(istream & inp) {in = &inp;}

	virtual bool IsOutput() {return false;}
};

//implements binary formatting
class SERIALIZER_BIN_OUTPUT : public SERIALIZER
{
private:
	ostream * out;
	bool IsBigEndian()
	{
		short word = 0x4321;
		if((*(char *)& word) != 0x21 )
			return true;
		else
			return false;
	}
	bool bigendian;
	void ByteSwap(unsigned char * b, int n)
	{
		register int i = 0;
		register int j = n-1;
		while (i<j)
		{
			std::swap(b[i], b[j]);
			i++, j--;
		}
	}

protected:
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
		if (type == START && TrackingPointers())
		{
			SerializeData(uid);
		}
		return uid;
	}
	virtual void SerializeData(int & i)
	{
		if (!bigendian)
		{
			int temp(i);
			ByteSwap(reinterpret_cast<unsigned char *>(&temp),sizeof(int));
			out->write(reinterpret_cast<char *>(&temp),sizeof(int));
		}
		else
			out->write(reinterpret_cast<char *>(&i),sizeof(int));
	}
	virtual void SerializeData(float & i)
	{
		if (!bigendian)
		{
			float temp(i);
			ByteSwap(reinterpret_cast<unsigned char *>(&temp),sizeof(float));
			out->write(reinterpret_cast<char *>(&temp),sizeof(float));
		}
		else
			out->write(reinterpret_cast<char *>(&i),sizeof(float));
	}
	virtual void SerializeData(double & i)
	{
		if (!bigendian)
		{
			double temp(i);
			ByteSwap(reinterpret_cast<unsigned char *>(&temp),sizeof(double));
			out->write(reinterpret_cast<char *>(&temp),sizeof(double));
		}
		else
			out->write(reinterpret_cast<char *>(&i),sizeof(double));
	}
	virtual void SerializeData(string & i)
	{
		//unsigned long long strlen = i.length();
		int strlen = i.length();
		SerializeData(strlen);
		out->write(i.c_str(),strlen*sizeof(char));
	}
	virtual void SerializeData(unsigned long long & i)
	{
		if (!bigendian)
		{
			unsigned long long temp(i);
			ByteSwap(reinterpret_cast<unsigned char *>(&temp),sizeof(unsigned long long));
			out->write(reinterpret_cast<char *>(&temp),sizeof(unsigned long long));
		}
		else
			out->write(reinterpret_cast<char *>(&i),sizeof(unsigned long long));
	}
public:
	SERIALIZER_BIN_OUTPUT() : out(NULL) {bigendian = IsBigEndian();}
	SERIALIZER_BIN_OUTPUT(ostream & newout) : out(&newout) {bigendian = IsBigEndian();}

	virtual void SetOutput(ostream & outp) {out = &outp;}

	virtual bool IsOutput() {return true;}
};

class SERIALIZER_BIN_INPUT : public SERIALIZER
{
private:
	istream * in;
	bool IsBigEndian()
	{
		short word = 0x4321;
		if((*(char *)& word) != 0x21 )
			return true;
		else
			return false;
	}
	bool bigendian;
	void ByteSwap(unsigned char * b, int n)
	{
		register int i = 0;
		register int j = n-1;
		while (i<j)
		{
			std::swap(b[i], b[j]);
			i++, j--;
		}
	}

protected:
	virtual void SerializeData(int & i)
	{
		in->read(reinterpret_cast<char *>(&i),sizeof(int));

		if (!bigendian)
		{
			ByteSwap(reinterpret_cast<unsigned char *>(&i),sizeof(int));
		}
	}
	virtual void SerializeData(float & i)
	{
		in->read(reinterpret_cast<char *>(&i),sizeof(float));

		if (!bigendian)
		{
			ByteSwap(reinterpret_cast<unsigned char *>(&i),sizeof(float));
		}
	}
	virtual void SerializeData(double & i)
	{
		in->read(reinterpret_cast<char *>(&i),sizeof(double));

		if (!bigendian)
		{
			ByteSwap(reinterpret_cast<unsigned char *>(&i),sizeof(double));
		}
	}
	virtual void SerializeData(string & i)
	{
		//unsigned long long strlen = 0;
		int strlen = 0;
		SerializeData(strlen);
		char inbuffer[strlen+1];
		inbuffer[strlen] = '\0';
		in->read(inbuffer,strlen*sizeof(char));
		assert(in->gcount() == strlen);
		i.clear();
		for (int n = 0; n < strlen; n++)
		{
			i.push_back(inbuffer[n]);
		}
	}
	virtual void SerializeData(unsigned long long & i)
	{
		in->read(reinterpret_cast<char *>(&i),sizeof(unsigned long long));

		if (!bigendian)
		{
			ByteSwap(reinterpret_cast<unsigned char *>(&i),sizeof(unsigned long long));
		}
	}
	virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
	{
		unsigned long long output_uid = uid;

		if (type == START && TrackingPointers())
		{
			SerializeData(output_uid);
		}

		return output_uid;
	}
public:
	SERIALIZER_BIN_INPUT() : in(NULL) {bigendian = IsBigEndian();}
	SERIALIZER_BIN_INPUT(istream & newin) : in(&newin) {bigendian = IsBigEndian();}

	virtual void SetInput(istream & inp) {in = &inp;}

	virtual bool IsOutput() {return false;}
};

//implements simpletree input/output
class SERIALIZER_TREE_VALUE
{
	public:
		string value;
		unsigned long long uid;
};

class SERIALIZER_TREE_OUTPUT : public SERIALIZER
{
	private:
		SIMPLETREE <string, SERIALIZER_TREE_VALUE> * tree;

	protected:
		virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
		{
			if (!tree)
			{
				//uh oh, throw some kind of error
				cerr << "Serialization Error: noticed tree is NULL when writing value " << name << endl;
			}

			if (type == START)
			{
				tree->branch[name].parent = tree;
				if (TrackingPointers())
				{
					tree->branch[name].value.uid = uid;
				}
				tree = &tree->branch[name];
			}
			else if (type == END)
			{
				if (!tree->parent)
				{
					//uh oh, throw some kind of error
					cerr << "Serialization Error: noticed tree is corrupted when writing " << name << endl;
				}
				else
				{
					tree = tree->parent;
				}
			}
			return uid;
		}
		virtual void SerializeData(int & i)
		{
			stringstream s;
			s << i;
			tree->value.value = s.str();
		}
		virtual void SerializeData(float & i)
		{
			stringstream s;
			s << std::scientific << std::setprecision (20) << i;
			tree->value.value = s.str();
		}
		virtual void SerializeData(double & i)
		{
			stringstream s;
			s << std::scientific << std::setprecision (20) << i;
			tree->value.value = s.str();
		}
		virtual void SerializeData(string & i)
		{
			tree->value.value = i;
		}
		virtual void SerializeData(unsigned long long & i)
		{
			stringstream s;
			s << i;
			tree->value.value = s.str();
		}
	public:
		SERIALIZER_TREE_OUTPUT() : tree(NULL) {}
		SERIALIZER_TREE_OUTPUT(SIMPLETREE <string, SERIALIZER_TREE_VALUE> & out) : tree(&out) {SetOutput(out);}

		virtual void SetOutput(SIMPLETREE <string, SERIALIZER_TREE_VALUE> & out) {tree = &out;tree->branch.clear();tree->parent = NULL;tree->value.value = "root";}

		virtual bool IsOutput() {return true;}
};

class SERIALIZER_TREE_INPUT : public SERIALIZER
{
	private:
		SIMPLETREE <string, SERIALIZER_TREE_VALUE> * tree;

	protected:
		virtual void SerializeData(int & i)
		{
			stringstream s;
			s.str(tree->value.value);
			s >> i;
		}
		virtual void SerializeData(float & i)
		{
			stringstream s;
			s.str(tree->value.value);
			s >> i;
		}
		virtual void SerializeData(double & i)
		{
			stringstream s;
			s.str(tree->value.value);
			s >> i;
		}
		virtual void SerializeData(string & i)
		{
			i = tree->value.value;
		}
		virtual void SerializeData(unsigned long long & i)
		{
			stringstream s;
			s.str(tree->value.value);
			s >> i;
		}
		virtual unsigned long long FormatInfo(string name, FORMAT_TYPE type, unsigned long long uid, bool builtin)
		{
			if (!tree)
			{
				//uh oh, throw some kind of error
				cerr << "Serialization Error: noticed tree is NULL when writing value " << name << endl;
			}

			unsigned long long output_uid = uid;

			if (type == START)
			{
				if (TrackingPointers())
				{
					output_uid = tree->branch[name].value.uid;
				}
				tree = &tree->branch[name];
			}
			else if (type == END)
			{
				if (!tree->parent)
				{
					//uh oh, throw some kind of error
					cerr << "Serialization Error: noticed tree is corrupted when writing " << name << endl;
				}
				else
				{
					tree = tree->parent;
				}
			}

			return output_uid;
		}
	public:
		SERIALIZER_TREE_INPUT() : tree(NULL) {}
		SERIALIZER_TREE_INPUT(SIMPLETREE <string, SERIALIZER_TREE_VALUE> & in) : tree(&in) {SetInput(in);}

		virtual void SetInput(SIMPLETREE <string, SERIALIZER_TREE_VALUE> & in)
		{
			tree = &in;
			if (tree->value.value != "root")
			{
				//uh oh, throw some kind of error
				cerr << "Serialization Error: tried to set input to corrupted tree: expected root, got " << tree->value.value << endl;
			}
		}

		virtual bool IsOutput() {return false;}
};

#define _SERIALIZE_(ser,varname) ser->Serialize(#varname,varname)
#define _STARTSERIALIZE_(ser,varname) ser->StartSerialize(#varname,varname)
#define _SERIALIZEPTR_(ser,varname) ser->SerializePointer(#varname,varname)
#define _SERIALIZELIST_(ser,varname) ser->SerializeList(#varname,varname)
#define _SERIALIZEVECTOR_(ser,varname) ser->SerializeVector(#varname,varname)
#define _SERIALIZEMAP_(ser,varname) ser->SerializeMap(#varname,varname)



//example usage:

/*//example POD class
class VERTEX
{
friend class SERIALIZER;
public:
	VERTEX() {}
	VERTEX(float nx, float ny, float nz) : x(nx),y(ny),z(nz) {}
	float x,y,z;
	bool operator==(const VERTEX & other) {return (memcmp(this,&other,sizeof(VERTEX)) == 0);}

	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s,x);
		_SERIALIZE_(s,y);
		_SERIALIZE_(s,z);
	}
};

//example complex class
class PLAYER
{
friend class SERIALIZER;
private:
	VERTEX position;
	int animframe;
	string name;
	list <int> simplelist;
	list <VERTEX> complexlist;
	VERTEX * curpos;
	VERTEX * nextpos;
	VERTEX * lastpos;

public:
	string & GetName() {return name;}
	PLAYER()
	{
		name = "Test player";
		position.x = position.y = position.z = 1.0;
		animframe = 1337;
		simplelist.push_back(9);
		simplelist.push_back(8);
		complexlist.push_back(VERTEX(7,6,5));
		complexlist.push_back(VERTEX(4,3,2));
		complexlist.push_back(VERTEX(1,0,-1));
		curpos = &position;
		nextpos = NULL;
		lastpos = &(complexlist.back());
	}

	void SetName(const string & newname) {name = newname;}

	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s,position);
		_SERIALIZE_(s,animframe);
		_SERIALIZE_(s,name);
		_SERIALIZEPTR_(s,curpos);
		_SERIALIZEPTR_(s,nextpos);
		_SERIALIZEPTR_(s,lastpos);
		_SERIALIZELIST_(s,simplelist);
		_SERIALIZELIST_(s,complexlist);
	}
};

int main()
{
	{
		PLAYER player1;

		SERIALIZER_XML_OUTPUT serialize_output(cout);
		_STARTSERIALIZE_((&serialize_output),player1);

		ofstream output_file("output.xml");
		serialize_output.SetOutput(output_file);
		serialize_output.Reset();
		_STARTSERIALIZE_((&serialize_output),player1);
		output_file.close();
	}

	cout << "------------" << endl;

	{
		PLAYER player1;
		ifstream input_file("output.xml");
		SERIALIZER_XML_INPUT serialize_input(input_file);
		_STARTSERIALIZE_((&serialize_input),player1);
		input_file.close();

		SERIALIZER_XML_OUTPUT serialize_output(cout);
		_STARTSERIALIZE_((&serialize_output),player1);
	}

	cout << "Main ends" << endl;

	return 0;
}*/

namespace TEST_FRAMEWORK
{
	class TEST_VERTEX
	{
	friend class ::SERIALIZER;
	public:
		TEST_VERTEX() : x(0), y(0), z(0) {}
		TEST_VERTEX(float nx, float ny, float nz) : x(nx),y(ny),z(nz) {}
		float x,y,z;
		bool operator==(const TEST_VERTEX & other) {return (memcmp(this,&other,sizeof(TEST_VERTEX)) == 0);}
		bool operator!=(const TEST_VERTEX & other) {return !operator==(other);}

		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,x);
			_SERIALIZE_(s,y);
			_SERIALIZE_(s,z);
		}
	};

	//example complex class
	class TEST_PLAYER
	{
		friend class ::SERIALIZER;
	//private:
	public: //the test needs access to these; otherwise they'd be private
		TEST_VERTEX position;
		int animframe;
		string name;
		string player_description;
		bool alive;
		bool aiming;
		list <int> simplelist;
		list <TEST_VERTEX> complexlist;
		TEST_VERTEX * curpos;
		TEST_VERTEX * nextpos;
		TEST_VERTEX * lastpos;
		map <string, TEST_VERTEX> mymap;
		vector <TEST_VERTEX> myvector;
		double curtime;

	public:
		string & GetName() {return name;}
		TEST_PLAYER(bool defaults)
		{
			if (defaults)
			{
				name = "Test player";
				position.x = position.y = position.z = 1.0;
				animframe = 1337;
				simplelist.push_back(9);
				simplelist.push_back(8);
				complexlist.push_back(TEST_VERTEX(7,6,5));
				complexlist.push_back(TEST_VERTEX(4,3,2));
				complexlist.push_back(TEST_VERTEX(1,0,-1));
				curpos = &position;
				nextpos = NULL;
				lastpos = &(complexlist.back());

				mymap["testing123"] = TEST_VERTEX(1,2,3);
				mymap["testing654"] = TEST_VERTEX(6,5,4);

				myvector.push_back(TEST_VERTEX(5,4,3));
				myvector.push_back(TEST_VERTEX(2,1,0));

				player_description = "Hello there.\nHow are you??";
				alive = true;
				aiming = false;

				curtime = 0.123456789;
			}
		}

		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,position);
			_SERIALIZE_(s,animframe);
			_SERIALIZE_(s,name);
			_SERIALIZEPTR_(s,curpos);
			_SERIALIZEPTR_(s,nextpos);
			_SERIALIZEPTR_(s,lastpos);
			_SERIALIZELIST_(s,simplelist);
			_SERIALIZELIST_(s,complexlist);
			_SERIALIZEMAP_(s,mymap);
			_SERIALIZEVECTOR_(s,myvector);
			_SERIALIZE_(s,player_description);
			_SERIALIZE_(s,alive);
			_SERIALIZE_(s,aiming);
			_SERIALIZE_(s,curtime);
		}
	};
};

#endif
