#include "ai.h"
#include "car.h"
#include "bezier.h"
#include "track.h"
#include "carinput.h"
#include "mathvector.h"
#include "carwheelposition.h"

#include <cassert>
#include <cmath>
#include <algorithm>
#include <iostream>

#define GRAVITY 9.8

//used to calculate brake value
#define MAX_SPEED_DIFF 6.0
#define MIN_SPEED_DIFF 1.0

//used to find the point the car should steer towards to
#define LOOKAHEAD_FACTOR1 2.25
#define LOOKAHEAD_FACTOR2 0.33

//used to detect very sharp corners like Circuit de Pau
#define LOOKAHEAD_MIN_RADIUS 8.0

//used to calculate friction coefficient
#define FRICTION_FACTOR_LONG 0.68
#define FRICTION_FACTOR_LAT 0.62

const std::vector <float> & AI::GetInputs(CAR * car) const
{
	for( std::vector<AI_Car>::const_iterator it = AI_Cars.begin (); it != AI_Cars.end (); it++ )
	{
		if (car == it->car)
			return it->inputs;
	}
	return empty_vector;
}

void AI::add_car(CAR * car)
{
	assert(car);
	AI_Cars.push_back(AI_Car(car));
	assert(car->GetTCSEnabled());
	assert(car->GetABSEnabled());
	car->SetAutoShift(true);
	car->SetAutoClutch(true);
}

void AI::update(float dt, TRACK* track_p)
{
	for( std::vector<AI_Car>::iterator it = AI_Cars.begin (); it != AI_Cars.end (); it++ )
	{
		updateGasBrake(&(*it), dt, track_p);
		updateSteer(&(*it), dt);
	}
}


void AI::updateGasBrake(AI_Car *c, float dt, TRACK* track_p)
{
	float brake_value = 0.0;
	float gas_value = 0.5;

	if (c->car->GetEngineRPM() < c->car->GetEngineStallRPM())
		c->inputs[CARINPUT::START_ENGINE] = 1.0;
	else
		c->inputs[CARINPUT::START_ENGINE] = 0.0;
	
	//std::cout << c->car->GetEngineRPM() << std::endl;

	/*if( game.state.GetGameState() == STATE_STAGING )
	{
		//have to calculate friction coefficient after the simulation start,
		//since it's dependent on the normal force (car weight)
		calcMu(c, track_p);
		c->car->brake(1.0,0.0);
		return;
	}*/
	//commented out the staging  stuff above since it's not implemented yet and moved the calcMu function here -VENZON
	calcMu(c, track_p);

	const BEZIER *curr_patch = c->car->GetCurPatch(0);
	//if car is not on track, just let it roll
        if (!curr_patch)
	{
		curr_patch = c->car->GetCurPatch(1); //let's try the other wheel
		if (!curr_patch)
		{
			c->inputs[CARINPUT::THROTTLE] = 0.8;
			c->inputs[CARINPUT::BRAKE] = 0.0;
			return;
		}
	}

	//maximum lookahead distance
	/*VERTEX patch_direction = ((curr_patch->points[0][0] + curr_patch->points[0][3]) -
				 (curr_patch->points[3][0] + curr_patch->points[3][3])).ScaleR(0.5);
	Vamos_Geometry::Three_Vector direction_vector;
	direction_vector[0] = patch_direction.x;
	direction_vector[2] = patch_direction.y;
	direction_vector[1] = -patch_direction.z;*/
	//MATHVECTOR <float, 3> next_position(dest_point[2], dest_point[0], dest_point[1]);
	
	MATHVECTOR <float, 3> direction_vector = ((curr_patch->GetPoint(0,0) + curr_patch->GetPoint(0,3)) -
			(curr_patch->GetPoint(3,0) + curr_patch->GetPoint(3,3))) * 0.5;
			
	MATHVECTOR <float, 3> patch_direction (direction_vector[2], direction_vector[0], direction_vector[1]);
	
	//this version uses the velocity along tangent vector. it should calculate a lower current speed,
	//hence higher gas value or lower brake value
	//float currentspeed = c->car->chassis().cm_velocity().component(direction_vector);
	float currentspeed = c->car->GetVelocity().dot(patch_direction.Normalize());
	//this version just uses the velocity, do not care about the direction
	//float currentspeed = c->car->chassis().cm_velocity().magnitude();

	//check speed against speed limit of current patch
	float speed_limit = calcSpeedLimit(c, curr_patch, c->lateral_mu);
	float speed_diff = speed_limit - currentspeed;

	if (speed_diff < 0.0)
	{
		if (-speed_diff < MIN_SPEED_DIFF) //no need to brake if diff is small
		{
			brake_value = 0.0;
		}
		else
		{
			brake_value = -speed_diff / MAX_SPEED_DIFF;
			if (brake_value > 1.0) brake_value = 1.0;
		}
		gas_value = 0.0;
	}
	else if (isnan(speed_diff) || speed_diff > MAX_SPEED_DIFF)
	{
		gas_value = 1.0;
		brake_value = 0.0;
	}
	else
	{
		gas_value = speed_diff / MAX_SPEED_DIFF;
		brake_value = 0.0;
	}

	//check upto maxlookahead distance
	float maxlookahead = calcBrakeDist(c, currentspeed, 0.0, c->longitude_mu);
	float dist_checked = 0.0;
	float brake_dist = 0.0;
	const BEZIER *patch_to_check = curr_patch;

	while (dist_checked < maxlookahead)
	{
		patch_to_check = patch_to_check->next_patch;
		//if there is no next patch(probably a non-closed track, just let it roll
		if (!patch_to_check)
		{
			brake_value = 0.0;
			dist_checked = maxlookahead;
			break;
		}
		speed_limit = calcSpeedLimit(c, patch_to_check, c->lateral_mu);
		//VERTEX center1 = (patch_to_check->points[0][0] + patch_to_check->points[0][3]);
		//VERTEX center2 = (patch_to_check->points[3][0] + patch_to_check->points[3][3]);
		MATHVECTOR <float, 3> center1 = (patch_to_check->GetPoint(0,0) + patch_to_check->GetPoint(0,3));
		MATHVECTOR <float, 3> center2 = (patch_to_check->GetPoint(3,0) + patch_to_check->GetPoint(3,3));
		dist_checked += (center1 - center2).Magnitude() * 0.5;
		brake_dist = calcBrakeDist(c, currentspeed, speed_limit, c->longitude_mu);
		if (brake_dist > dist_checked)
		{
			brake_value = 1.0;
			gas_value = 0.0;
		}
	}

	if (c->car->GetGear() == 0 && !c->car->Shifting())
	{
		c->inputs[CARINPUT::SHIFT_UP] = 1.0;
		gas_value = 0.2;
		//std::cout << "UPSHIFT" << std::endl;
	}
	else
		c->inputs[CARINPUT::SHIFT_UP] = 0.0;
	//c->car->brake(brake_value,0.0);
	//c->car->gas(gas_value,0.0);
	c->inputs[CARINPUT::THROTTLE] = gas_value;
	c->inputs[CARINPUT::BRAKE] = brake_value;
	
	//if (-speed_diff > MIN_SPEED_DIFF) assert(gas_value == 0.0);
	
	//if (speed_limit < 1e6) assert(speed_diff < 1e6);
	/*std::cout << "patch direction: " << patch_direction << std::endl;
	//std::cout << ": " <<  << std::endl;
	std::cout << "velocity: " << c->car->GetVelocity() << std::endl;
	std::cout << "currentspeed: " << currentspeed << std::endl;
	//std::cout << "maxlookahead: " << maxlookahead << std::endl;
	std::cout << "speed_limit: " << speed_limit << std::endl;
	std::cout << "speed_diff: " << speed_diff << std::endl;
	std::cout << "gas,brake: " << gas_value << ", " << brake_value << std::endl;
	std::cout << std::endl;*/
	//assert(c->car->GetGear() == 0);
}

void AI::calcMu(AI_Car *c, TRACK* track_p)
{
	int i;
	float long_friction = 0.0;
	float lat_friction = 0.0;

	//float friction1(1.0), friction2(1.0); //commented out because in the refactored code the track always has 1.0 friction -VENZON
	//track_p->GetFrictionParams(friction1,friction2);

	for (i=0; i<4; i++)
	{
		//Vamos_Body::Wheel* wheel = c->car->wheel(i);
		//double sp = wheel->GetTire()->GetSigmaHat(); //ideal slip ratio given tire loading 
		//double maxfx = wheel->GetTire()->GetFx(sp); //maximum tire friction given ideal slip ratio
		//maxfx *= (1.0 - wheel->GetTire()->tread()) * friction1 + wheel->GetTire()->tread() * friction2;
		long_friction += c->car->GetTireMaxFx(WHEEL_POSITION(i));
		//double sa = wheel->GetTire()->GetAlphaHat(); //ideal slip angle given tire loading 
		//double maxfy = wheel->GetTire()->GetFy(sa); //maximum tire friction given ideal slip angle
		//maxfy *= (1.0 - wheel->GetTire()->tread()) * friction1 + wheel->GetTire()->tread() * friction2;
		lat_friction += c->car->GetTireMaxFy(WHEEL_POSITION(i));
		//lat_friction += maxfy;
	}

	float long_mu = FRICTION_FACTOR_LONG * long_friction / c->car->GetMass() / GRAVITY;
	float lat_mu = FRICTION_FACTOR_LAT * lat_friction / c->car->GetMass() / GRAVITY;
	if (!isnan(long_mu)) c->longitude_mu = long_mu;
	if (!isnan(lat_mu)) c->lateral_mu = lat_mu;
}

float AI::calcSpeedLimit(AI_Car *c, const BEZIER* patch, float friction)
{
	//adjust the radius at corner exit to allow a higher speed.
	//this will get the car to accellerate out of corner
	MATHVECTOR <float, 3> center1 = (patch->GetPoint(0,0) + patch->GetPoint(3,0));
	MATHVECTOR <float, 3> center2 = (patch->GetPoint(0,3) + patch->GetPoint(3,3));
	double track_width = (center1 - center2).Magnitude() * 0.5;
	double adjusted_radius = patch->track_radius;
	BEZIER* next_patch = patch->next_patch;
	
	if (next_patch && next_patch->track_radius > patch->track_radius &&
	    patch->track_radius > LOOKAHEAD_MIN_RADIUS)
	{
		adjusted_radius = patch->track_radius + track_width; 
	}
	
	//no downforce
	//float v1 = sqrt(friction * GRAVITY * adjusted_radius);

	//take into account downforce
	double denom = (1.0 - std::min(1.01, adjusted_radius * -(c->car->GetAerodynamicDownforceCoefficient()) * friction / c->car->GetMass()));
	double real = (friction * GRAVITY * adjusted_radius) / denom;
	double v2 = 1000.0; //some really big number
	if (real > 0)
		v2 = sqrt(real);
	
	return v2;
}

float AI::calcBrakeDist(AI_Car *ai_car, float current_speed, float allowed_speed, float friction)
{
	float c = friction * GRAVITY;
	float d = (-(ai_car->car->GetAerodynamicDownforceCoefficient()) * friction + ai_car->car->GetAeordynamicDragCoefficient())
			 / ai_car->car->GetMass();
	float v1sqr = current_speed * current_speed;
	float v2sqr = allowed_speed * allowed_speed;
	return -log((c + v2sqr*d)/(c + v1sqr*d))/(2.0*d);
}

/*void AI::updateGear(AI_Car *c, float dt)
{
	if ((c->car->ShiftPending()) && (c->car->gear() == 0))
        {
        	c->car->disengage_clutch (0.2);
                c->car->shift(1);
                c->shift_time = AUTO_CLUTCH_ENGAGE_TIME;
        }

	int current_gear = c->car->gear();
	// only autoshift when no shift pending and we are not in neutral or reverse gear
	if( !c->car->ShiftPending() && ( current_gear > 0 ) )
	{
		double peak_engine_speed = c->car->engine()->peak_engine_speed();
		// shift up when engine speed exceeds peak speed
		if( ( c->car->engine()->rotational_speed() > peak_engine_speed ) &&
		    ( c->car->engine()->last_rotational_speed() <= peak_engine_speed ) &&
		    ( current_gear < c->car->transmission()->forward_gears() ) )
		{
        		c->car->disengage_clutch (0.2);
	                c->car->shift_up();
        	        c->shift_time = AUTO_CLUTCH_ENGAGE_TIME;
		}

		// shift down when engine speed below shift_down_point
		// however, we do not auto shift down from 1st gear to neutral
		if( ( current_gear > 1 ) && 
		    c->car->engine()->rotational_speed() < 
		    DownshiftPoint( c, current_gear ) )
		{
        		c->car->disengage_clutch (0.2);
	                c->car->shift_down();
        	        c->shift_time = AUTO_CLUTCH_ENGAGE_TIME;
		}
	}
}

float AI::DownshiftPoint(AI_Car *c, int gear)
{
	float shift_down_point = 200.0;
	if( gear > 2 )
	{
		double current_gear_ratio = c->car->transmission()->gear_ratio( gear );
		double lower_gear_ratio = c->car->transmission()->gear_ratio( gear - 1 );
		float peak_engine_speed = c->car->engine()->peak_engine_speed();
		shift_down_point = peak_engine_speed / lower_gear_ratio * current_gear_ratio;
		shift_down_point -= ( 200.0 / gear );
	}
	return shift_down_point;
}

void AI::updateClutch(AI_Car *c, float dt)
{
	float autoclutch = 0.0;
        if (c->car->engine()->rotational_speed() < AUTO_CLUTCH_THRESH)
        {
        	float rotspeed = c->car->engine()->rotational_speed();
        	float stall = c->car->engine()->stall_speed() + AUTO_CLUTCH_MARGIN;
        	float clutch = (rotspeed-stall) / (AUTO_CLUTCH_THRESH-stall);
                if (clutch < 0)
                	clutch = 0;
                clutch = 1.0f - clutch;
                if (clutch > autoclutch)
                autoclutch = clutch;
        }

        if (c->shift_time > 0)
        {
                float clutch = c->shift_time / AUTO_CLUTCH_ENGAGE_TIME;
                if (clutch > autoclutch)
                        autoclutch = clutch;

                c->shift_time -= timefactor/fps;
        }

        if (!c->car->ShiftPending())
        {
                c->car->clutch (1.0f-autoclutch, 0.0);
        }
}*/

void AI::updateSteer(AI_Car *c, float dt)
{
	const BEZIER *curr_patch = c->car->GetCurPatch(0);
	//if car has no contact with track, just let it roll
	if (!curr_patch)
	{
		curr_patch = c->car->GetCurPatch(1); //let's try the other wheel
		if (!curr_patch)
		{
			if (!c->last_patch) return;
			//if car is off track, steer the car towards the last patch it was on
			//this should get the car back on track
			else curr_patch = c->last_patch;
		}
	}

	c->last_patch = curr_patch; //store the last patch car was on

	const BEZIER *next_patch = curr_patch->next_patch;
	//if there is no next patch (probably a non-closed track), let it roll
	if (!next_patch) return;

	//find the point to steer towards to
	MATHVECTOR <float, 3> center1 = (curr_patch->GetPoint(0,0) + curr_patch->GetPoint(3,0)); //left side of patch, center*2
	MATHVECTOR <float, 3> center2 = (curr_patch->GetPoint(0,3) + curr_patch->GetPoint(3,3)); //right side of patch, center*2
	float track_width = (center1 - center2).Magnitude() * 0.5;
	float lookahead = track_width * LOOKAHEAD_FACTOR1 + 
			c->car->GetVelocity().Magnitude() * LOOKAHEAD_FACTOR2;
	//std::cout << "lookahead: " << lookahead << std::endl;
	float length = 0.0;
	MATHVECTOR <float, 3> dest_point;
	while (length < lookahead)
	{
		/*MATHVECTOR <float, 3> center1 = (next_patch->GetPoint(0,0) + next_patch->GetPoint(3,0));
		MATHVECTOR <float, 3> center2 = (next_patch->GetPoint(0,3) + next_patch->GetPoint(3,3));
		length += (center1 - center2).Magnitude();
		dest_point = center1 * 0.5;
		next_patch = next_patch->next_patch;*/
		
		MATHVECTOR <float, 3> center1 = (next_patch->GetPoint(0,0) + next_patch->GetPoint(0,3));
		MATHVECTOR <float, 3> center2 = (next_patch->GetPoint(3,0) + next_patch->GetPoint(3,3));
		length += (center1 - center2).Magnitude();
		dest_point = center1 * 0.5;
		next_patch = next_patch->next_patch;
		
		//if there is no next patch for whatever reason, stop lookahead
		//if next patch is a very sharp corner, stop lookahead
		if ((!next_patch) || next_patch->track_radius < LOOKAHEAD_MIN_RADIUS)
		{
			length = lookahead;
			break;
		}
	}

	//introduce some bias depends on whether it's left or right turn and the radius
	//float bias = 0.0;
	//if (next_patch->track_radius > 300.0) bias = 0.0;
	//else bias = 0.3 * (1 - next_patch->track_radius/300.0) * next_patch->turn;
	//VERTEX next_patch_center = next_patch->SurfCoord(0.5 + bias,0.5);

	/*Vamos_Geometry::Three_Vector next_position;
	next_position[0] = dest_point.x;
	next_position[2] = dest_point.y;
	next_position[1] = -dest_point.z;*/
	MATHVECTOR <float, 3> next_position(dest_point[2], dest_point[0], dest_point[1]);

	MATHVECTOR <float, 3> car_position = c->car->GetCenterOfMassPosition();
	MATHVECTOR <float, 3> car_orientation(0,1,0);
	QUATERNION <float> fixer;
	fixer.Rotate(-3.141593*0.5, 0, 0, 1);
	(c->car->GetOrientation()*fixer).RotateVector(car_orientation);
	//std::swap(car_orientation[0], car_orientation[1]);
	//car_orientation[0] = -car_orientation[0];

	MATHVECTOR <float, 3> desire_orientation = next_position - car_position;

	//car's direction on the horizontal plane
	car_orientation[2] = 0;
	//desired direction on the horizontal plane
	desire_orientation[2] = 0;
	
	car_orientation = car_orientation.Normalize();
	desire_orientation = desire_orientation.Normalize();

	//the angle between car's direction and unit y vector (forward direction)
	/*MATHVECTOR <float, 3> unit_y(1,0,0);
	double dotprod = car_orientation.dot(unit_y);
	assert(dotprod > -1.0 && dotprod < 1.0);
	double alpha = acos(dotprod);
	if (car_orientation[0] < 0) alpha = -alpha;
	alpha = alpha * 180.0 / M_PI;*/
	double alpha = Angle(car_orientation[0], car_orientation[1]);

	//the angle between desired direction and unit y vector (forward direction)
	/*dotprod = desire_orientation.dot(unit_y);
	assert(dotprod > -1.0 && dotprod < 1.0);
        double beta = acos(dotprod);
        if (desire_orientation[0] < 0) beta = -beta;
	beta = beta * 180.0 / M_PI;*/
	double beta = Angle(desire_orientation[0], desire_orientation[1]);

	//calculate steering angle and direction
	double angle = beta - alpha;
	if (angle > -360.0 && angle <= -180.0)
		angle = -(360.0 + angle);
	else if (angle > -180.0 && angle <= 0.0)
		angle = - angle;
	else if (angle > 0.0 && angle <= 180.0)
		angle = - angle;
	else if (angle > 180.0 && angle <= 360.0)
		angle = 360.0 - angle;

	float steer_value = angle / c->car->GetMaxSteeringAngle();
	if (steer_value > 1.0) steer_value = 1.0;
	else if (steer_value < -1.0) steer_value = -1.0;

	assert(!isnan(steer_value));
	c->inputs[CARINPUT::STEER_RIGHT] = steer_value;
	
	/*std::cout << "car: " << car_position << std::endl;
	std::cout << "next: " << next_position << std::endl;
	std::cout << "desired orient: " << desire_orientation << std::endl;
	std::cout << "car orient: " << car_orientation << std::endl;
	std::cout << alpha << ", " << beta << " = Steer: " << steer_value << std::endl;
	std::cout << std::endl;*/
}

double AI::Angle(double x1, double y1)
{
	return atan2(y1, x1) * 180.0 / M_PI;
}
